//
// HPDisk - Emulate a HP9121 diskette drive, using an SD card
//
// 20131103 ag	0.1	First working version
// 20131120 ag	0.2	Rearranged pins so that we can interrupt on change
// 20131124 ag	0.3	Changed so that we now write 256-byte sectors, two per real sector. Thus the resulting file is a true image of the LIF "disk"
// 20131125 ag	0.4 Small tweaks. Does now work properly with SDEBUG undefined. Needed a small delay in the loop
// 20131208 ag	0.5 Corrected the reported status for write protect, added handling of Amigo Wakeup, added support for hardware PP
//
// Comments on bugs and suggested improvements are welcome. Mail me at gpib@dalton.ax
//
// Project page: http://www.dalton.ax/hpdisk
//
// Todo:
// Status 2 on SD change
// Zero trck etc on clear
#include <p18f4620.h>
#include <delays.h>
#include <stdio.h>
#include <stdlib.h>
#include "gpib.h"
#include "amigo.h"
#include "sio.h"

// States for the SD-open state machine
//
#define SD_MOUNT	1
#define SD_OPEN 	2
#define	SD_DONE		0


void format_disk(void);
void write_sectors(void);
void DumpData(unsigned char *Ptr,int length);
int SD_open(void);

#pragma config WDT = OFF        //Watchdog timer OFF
#pragma config PBADEN = OFF 	//PORTB 0 to 4 digital - not analogue
#pragma config OSC = HSPLL      //HSPLL oscillator, 8MHz crystal with PLL is 32MHz clock
#pragma config LVP = OFF        //Single supply ICSP OFF
#pragma config FCMEN = OFF      //Fail safe clock monitor OFF
#pragma config IESO = OFF       //
#pragma config PWRT = OFF
#pragma config BOREN = OFF
#pragma config MCLRE = ON
#pragma config STVREN = ON		// Stack full/underflow will cause Reset  


#include "Globals.h"
#include "SD.h"
#include "pff.h"
//
// Initialized data
//
#pragma idata myidata
char file[] = "/lifdata.bin"; 	//File name to open

// Uninitialized data
//
#pragma udata
FATFS fs; // File system object
FRESULT rc;

extern unsigned char string[MAX_DBG_STRING];
extern unsigned char edge[5];

unsigned char MMC; 				// Memory card status
DWORD fpos;						// File position

#pragma udata my_section_2
char buffer[512];
//
// You need to pre-create the LIF image file, liftata.bin. Use this from DOS for example
//
// echo "This is just a sample line appended to create a big file " > lifdata.bin
//for /L %i in (1,1,14) do type lifdata.bin >> lifdata.bin
//
void main(void)
{
    char Key = 0;
	unsigned int i = 0;
	int nbytes;
	unsigned char status;
	//
	LED1_TRIS = 0;
	LED2_TRIS = 0;
	//
	LED1_PORT = 1;
	//
	//memset(edge,0,5);
	edge[0]=0;
	edge[1]=0;
	edge[2]=0;
	edge[3]=0;
    Globals_IO_init();  	// Set in and outputs for SD-Card
	gpib_init();			// Init GPIB
	sio_setup();			// Init serial port for debugging

	sprintf(string,(const far rom char*) "\r\n\r\nHPDisk v 0.5 %S\r\n", __DATE__);
	sio_puts(string);
	delay_ms(500);
	//sprintf(string,(const far rom char*)"\r\nStart %d %d", CD_PORT, WP_PORT);
	//sio_puts(string);
	//
	LED2_PORT = 1; 
	//format_disk();

	status = 0;

	while(1)				// Main loop, forever
	{
		rc=SD_open();
		sprintf(string,(const far rom char*)"\r\n\r\nSD_open returns %d", rc);
		sio_puts(string);
		if(rc != 0) continue;
		//
		if(WP_PORT)
		{
			sprintf(string,(const far rom char*)"\r\n\Card is write protected");
			sio_puts(string);
		}
		//
		// Strategy: Scan GPIB continously. readgpib() returns zero if nothing or a command byte, corresponding to an Amigo Command
		// While reading, readgpib() pushes every byte to the Amigo-routines, a valid command breaks out of the receive loop
		// All command handling is done in the Amigo-routines, with two exceptions: Ident and parallell poll
		//
		amigo_init();
		while(CD_PORT == 0) // SDCard CD is low when card present, so loop here and fall through if card removed (PORTCbits.RC1)
		{
			gpib_ptr = 0;
			
			LED2_PORT = !LED2_PORT;
			status = readgpib();
			if((status && status != 255) || gpib_ptr>0)
			{

				sprintf(string,(const far rom char*)" - Status: %d  \r\n", status);
				sio_puts(string);

#if defined(SDEBUG)
				sio_puts(string);
				DumpData(gpib_buff, gpib_ptr);
#else
				delay_ms(10);			// Found out the hard way... We really need a short delay here (disks are slow)
#endif
			}
			//
			if(status == AMIGO_CMD_CLEAR) amigo_cmd_clear();
			if(status == AMIGO_CMD_SEND_STATUS) amigo_cmd_send_status();
			if(status == AMIGO_CMD_SEEK) amigo_cmd_seek();
			if(status == AMIGO_CMD_BUFFERED_READ) amigo_cmd_buffered_read();
			if(status == AMIGO_CMD_DSJ) amigo_cmd_dsj();
			if(status == AMIGO_CMD_BUFFERED_WRITE) amigo_cmd_buffered_write();
			if(status == AMIGO_CMD_WAKEUP) amigo_cmd_wakeup();
			//
			if(edge[0]!=0)
			{
				sprintf(string,(const far rom char*)"\r\nEdge: %s  %d", edge, edge[0]);
				sio_puts(string);
			}	
		}
		sprintf(string,(const far rom char*)"\r\n\r\nOops. No card");
		sio_puts(string);
		LED1_PORT = 0; 
		delay_ms(1000);
		LED1_PORT = 1; 
		delay_ms(1000);
	}

}
//
// For testing. Formats the file with valid LIF-contents
//
void write_sectors(void)
{
DWORD fpos, sector;
int nbytes, temp,i;

FRESULT rc;
	for(sector=0;sector < (16L * 2L );sector++) // * 66L
	{
		
		//sprintf(string,(const far rom char*)"\r\nWrite sector: %ld %X ", sector, fpos);
		//sio_puts((unsigned char*)string);
		if(sector % 2 == 0)	// even
		{
			// Read the upper part of the 512-byte sector, ie sector+1
			fpos = (sector / 2) * 512L + 256L;
			rc=pf_lseek(fpos);
			rc=pf_read(buffer,256,&nbytes);
			sprintf(string,(const far rom char*)"\r\nRead odd sector: %ld %lX %d", sector, fpos, rc);
			sio_puts((unsigned char*)string);
			// Seek to start of sector
			fpos = sector / 2 * 512L;
			rc=pf_lseek(fpos);
			sprintf(string,(const far rom char*)"\r\nWrite even sector: %ld %lX %d", sector, fpos, rc);
			sio_puts((unsigned char*)string);
			// Write data
			temp = strlen(string);
			rc=pf_write(string, temp, &nbytes);
			for(i=0;i<256-temp;i++)
				rc=pf_write(&string[8], 1, &nbytes);
			// Write upper part
			rc=pf_write(buffer, 256, &nbytes);
			// finalise
			rc =pf_write(0, 0, &nbytes);
		}
		else			// odd
		{

			// Read the lower part of the 512-byte sector, ie sector
			fpos = sector / 2 * 512L;
			rc=pf_lseek(fpos);
			rc=pf_read(buffer,256,&nbytes);
			sprintf(string,(const far rom char*)"\r\nRead even sector: %ld %lX %d", sector, fpos, rc);
			sio_puts((unsigned char*)string);
			// Write lower part
			rc=pf_write(buffer, 256, &nbytes);

			fpos = (sector / 2) * 512L + 256L;
			sprintf(string,(const far rom char*)"\r\nWrite odd sector: %ld %lX ", sector, fpos);
			sio_puts((unsigned char*)string);
			// Write data
			temp = strlen(string);
			rc=pf_write(string, temp, &nbytes);
			for(i=0;i<256-temp;i++)
				rc=pf_write(string[8], 1, &nbytes);

			// finalise
			rc =pf_write(0, 0, &nbytes);
		}
	}
}
void format_disk(void)
{
int nbytes;
	rc=0;
	sprintf(string,(const far rom char*)"\r\n\r\nMounting");
	sio_puts(string);
	delay_ms(500);
    rc=pf_mount(&fs); //Mount volume
	sprintf(string,(const far rom char*)"\r\nMount returns %X", rc);
	sio_puts(string);
    MMC = pf_open(file); //Open the file
	sprintf(string,(const far rom char*)"\r\nOpen returns %X", MMC);
	sio_puts(string);
	sprintf(string,(const far rom char*)"\r\nReading");
	sio_puts(string);
	memset(buffer,0,10);
    rc=pf_read(buffer,10,&nbytes);
	sprintf(string,(const far rom char*)"\r\nRead returns %X, %d bytes read: %s", rc, nbytes, buffer);
	sio_puts(string);
	//
	fpos = 0;
	memset(buffer,' ',511);
	memcpy(buffer,&vl,sizeof(vl));	// Volume 
	rc=pf_lseek(fpos);
	if(rc)
	{
		sprintf(string,(const far rom char*)"\r\nSeek fail %d", rc);
		sio_puts(string);
	}
	rc=pf_write(buffer, 512, &nbytes);
	if(rc)
	{
		sprintf(string,(const far rom char*)"\r\nWrite fail %d", rc);
		sio_puts(string);
	}
	pf_write(0, 0, &nbytes);
	if(rc)
	{
		sprintf(string,(const far rom char*)"\r\nWrite fail %d", rc);
		sio_puts(string);
	}
	fpos = (DWORD)2*512L;
	memset(buffer,' ',511);
	de.filetype = 0x100;
	memcpy(buffer,&de,sizeof(de));
	memcpy(buffer+sizeof(de),&de,sizeof(de));
	de.filetype = 0xffff;
	memcpy(buffer+sizeof(de)+sizeof(de),&de,sizeof(de));
	rc=pf_lseek(fpos);
	if(rc)
	{
		sprintf(string,(const far rom char*)"\r\nSeek fail %d", rc);
		sio_puts(string);
	}
	rc=pf_write(buffer, 512, &nbytes);
	if(rc)
	{
		sprintf(string,(const far rom char*)"\r\nWrite fail %d", rc);
		sio_puts(string);
	}
	pf_write(0, 0, &nbytes);
	if(rc)
	{
		sprintf(string,(const far rom char*)"\r\nWrite fail %d", rc);
		sio_puts(string);
	}
	rc=pf_mount(0); //unmount volume
	sprintf(string,(const far rom char*)"\r\nDone");
	sio_puts(string);
    while (1)
	;
}
//
// Dump an data-area in human-readable form
//
void DumpData(unsigned char *Ptr,int length){
int i,j;
	sprintf(string,(const far rom char*)"\r\nDump %X %X\r\n",Ptr, length);
	sio_puts(string);
	for(j=0;j<80&&(j*16<=length);j++){
		sprintf(string,(const far rom char*)"\r\n");
		sio_puts(string);
		for(i=0;i<16 && (i+j*16<=length);i++)
			{
				sprintf(string, " %02X",(unsigned char) *(Ptr+i+j*16));
				sio_puts(string);
			}
		sprintf(string, " | ");
		sio_puts(string);
		for(i=0;i<16 && (i+j*16<=length);i++)
			{
				if(*(Ptr+i+j*16)>' ') sprintf(string, "%c",(*(Ptr+i+j*16))); 
				else sprintf(string, ".");
				sio_puts(string);
			}
	}
	sprintf(string, ">\r\n");
	sio_puts(string);
}
//
// Prepare the SD card and file for access. Do max 5 retries on each op. 
// Returns 0 on success or the actual state if failed
//
int SD_open()
{
int sd_state, retries;

	sd_state = SD_MOUNT;
	retries = 5;
	rc=0;
	while(retries > 0 && sd_state != SD_DONE)
	{
		--retries;
		switch(sd_state)
		{
			case SD_MOUNT:
#if defined(SDEBUG)
				sprintf(string,(const far rom char*)"\r\n\r\nMounting");
				sio_puts(string);
				delay_ms(50);	// Since mount will mess with the clock
#endif
    			rc=pf_mount(&fs); 		//Mount volume
#if defined(SDEBUG)
				sprintf(string,(const far rom char*)"\r\nMount returns %X", rc);
				sio_puts(string);
#endif
				if(rc == 0) sd_state = SD_OPEN;
				else delay_ms(500);
				break;
			case SD_OPEN:
				retries = 5;
				delay_ms(50);
    			MMC = pf_open(file); //Open the file
#if defined(SDEBUG)
				sprintf(string,(const far rom char*)"\r\nOpen %s returns %X", file, MMC);
				sio_puts(string);
#endif
				if(rc == 0) sd_state = SD_DONE;
				else delay_ms(500);
				break;
			case SD_DONE:
				break;
			default:
				break;
		}
	} // while
	return sd_state;
}